/*  This is the Cube Analyzer Visualization Plug-in for XMMS.
 *
 *  It displays a spectrum analyzer on the LED cube that can be
 *  connected to the serial port of your computer. Have a look at
 *  http://www.entropia.de/wiki/LED-Cube
 *
 *  It is based on the code of the sanalyzer visualization plug-in
 *  that comes with the XMMS source code package; some code from other
 *  plug-ins of XMMS may shamelessly be taken as well.
 *
 *  Copyright (C) 2004 of the changes from sanalyzer to Cube Analyzer
 *  by Max-Gerd Retzlaff <m.retzlaff@gmx.net>, http://bl0rg.net/~mgr/
 *
 *  As the code of sanalyzer has been licensed under the terms of the GPL,
 *  the code of this file is GPLed as well.
 *
 * 
 *  The following is the primordial starting comment of the sanalyzer code:
 */
/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas,
 *                           Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <gtk/gtk.h>
#include <math.h>

#include <xmms/plugin.h>
#include <xmms/util.h>
#include "xmms_logo.xpm"
#include "i18n.h"

#define WIDTH 135 // 250
#define HEIGHT 100
#define NUM_STEPS 6 // HEIGHT // 6 not 5! We have 0 LEDs on, 1 LED on, .., 5 LEDs on => 6 states!
#define NUM_BANDS 5 // 16
static const gint xscale[] = // {0, 3, 16,  30, 80, 255};
			     {0,1,4,15,47,255};
                             // {0, 3, 14,  40, 101, 255}; // <- distribution is a quick hack without much thought
                             // {0, 1, 2, 3, 5, 7, 10, 14, 20, 28, 40, 54, 74, 101, 137, 187, 255};
static gdouble yscale;

#define TTY_DELAY_LOOPS 5
/* 4 results in about 28-29 frames per second on my current configuration (IBM T42p)
 * To determine the framerate select an ordinary file to be output device file,
 * and fire up the following command (Bash syntax):
 * prev=0;while (true) ; do now=$(wc -l ttyUSB0 | awk '{print $1}'); let diff=$now-$prev; echo "$prev - $now = $diff";prev=$now ; sleep 1 ; done
 */
static gint tty_delay_counter;

static char cube[5][26]; // 5 plains with 25 LEDs each

/* The permutation arrays are created with some nifty Common Lisp code... :o) */
static const gint permutation[] = 
  {0, 25, 50, 75, 100, 1, 26, 51, 76, 101, 2, 27, 52, 77, 102, 3, 28, 53, 78, 103, 4, 29, 54, 79, 104, 5, 30, 55, 80, 105, 6, 31, 56, 81, 106, 7, 32, 57, 82, 107, 8, 33, 58, 83, 108, 9, 34, 59, 84, 109, 10, 35, 60, 85, 110, 11, 36, 61, 86, 111, 12, 37, 62, 87, 112, 13, 38, 63, 88, 113, 14, 39, 64, 89, 114, 15, 40, 65, 90, 115, 16, 41, 66, 91, 116, 17, 42, 67, 92, 117, 18, 43, 68, 93, 118, 19, 44, 69, 94, 119, 20, 45, 70, 95, 120, 21, 46, 71, 96, 121, 22, 47, 72, 97, 122, 23, 48, 73, 98, 123, 24, 49, 74, 99, 124};
// {0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100, 105, 110, 115, 120, 1, 6, 11, 16, 21, 26, 31, 36, 41, 46, 51, 56, 61, 66, 71, 76, 81, 86, 91, 96, 101, 106, 111, 116, 121, 2, 7, 12, 17, 22, 27, 32, 37, 42, 47, 52, 57, 62, 67, 72, 77, 82, 87, 92, 97, 102, 107, 112, 117, 122, 3, 8, 13, 18, 23, 28, 33, 38, 43, 48, 53, 58, 63, 68, 73, 78, 83, 88, 93, 98, 103, 108, 113, 118, 123, 4, 9, 14, 19, 24, 29, 34, 39, 44, 49, 54, 59, 64, 69, 74, 79, 84, 89, 94, 99, 104, 109, 114, 119, 124};
// {0, 1, 2, 3, 4, 25, 26, 27, 28, 29, 50, 51, 52, 53, 54, 75, 76, 77, 78, 79, 100, 101, 102, 103, 104, 5, 6, 7, 8, 9, 30, 31, 32, 33, 34, 55, 56, 57, 58, 59, 80, 81, 82, 83, 84, 105, 106, 107, 108, 109, 10, 11, 12, 13, 14, 35, 36, 37, 38, 39, 60, 61, 62, 63, 64, 85, 86, 87, 88, 89, 110, 111, 112, 113, 114, 15, 16, 17, 18, 19, 40, 41, 42, 43, 44, 65, 66, 67, 68, 69, 90, 91, 92, 93, 94, 115, 116, 117, 118, 119, 20, 21, 22, 23, 24, 45, 46, 47, 48, 49, 70, 71, 72, 73, 74, 95, 96, 97, 98, 99, 120, 121, 122, 123, 124};

static GtkWidget *window = NULL,*area;
static GdkPixmap *bg_pixmap = NULL, *draw_pixmap = NULL, *bar = NULL;
static GdkGC *gc = NULL;
static gint16 bar_heights[NUM_BANDS];
static gint timeout_tag;

static void cube_analyzer_init(void);
static void cube_analyzer_cleanup(void);
static void cube_analyzer_playback_start(void);
static void cube_analyzer_playback_stop(void);
static void cube_analyzer_render_freq(gint16 data[2][256]);

void cube_analyzer_about(void);

/* configure - BEGIN */
#include <stdio.h>
#include <xmms/configfile.h>

static GtkWidget *configure_win = NULL, *path_entry;

static gchar *ttyfile_path = NULL;
static FILE *ttyfile = NULL;

static void cube_analyzer_configure(void);
/* configure - END */

VisPlugin cube_analyzer_vp =
{
	NULL,
	NULL,
	0,
	NULL, /* Description */
	0, /* Numbers of PCM channels wanted in the call to render_pcm */
	1, /* Numbers of freq channels wanted in the call to render_freq */	
	cube_analyzer_init, /* init */
	cube_analyzer_cleanup, /* cleanup */
	cube_analyzer_about, /* about */
	cube_analyzer_configure, /* configure */
	NULL, /* disable_plugin */
	cube_analyzer_playback_start, /* playback_start */
	cube_analyzer_playback_stop, /* playback_stop */
	NULL, /* render_pcm */
	cube_analyzer_render_freq  /* render_freq */
};

VisPlugin *get_vplugin_info(void)
{
  
        cube_analyzer_vp.description =
		g_strdup_printf(_("LED-Cube Analyzer %s"), VERSION);
	return &cube_analyzer_vp;
}


static void cube_analyzer_destroy_cb(GtkWidget *w,gpointer data)
{
	cube_analyzer_vp.disable_plugin(&cube_analyzer_vp);
}

static void cube_analyzer_init(void)
{
	gint i, j;
        tty_delay_counter = 0;
        /* configure - begin */
	ConfigFile *cfgfile;

	/* init cube array */
	for(j = 0; j < 5; j++) {
		for(i = 0; i < 25; i++)
			cube[j][i] = '0';
		cube[j][25] = '\0';
	}


	cfgfile = xmms_cfg_open_default_file();
	if (cfgfile)
	{
		xmms_cfg_read_string(cfgfile, "cube_analyzer", "ttyfile_path", &ttyfile_path);
		xmms_cfg_free(cfgfile);
	}

	if (!ttyfile_path)
		ttyfile_path = g_strdup(g_get_home_dir());
	/* configure - end */

	GdkColor color;
	// int i;
	if(window)
		return;
	window = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title(GTK_WINDOW(window),_("Cube Spectrum analyzer"));
	gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);
	gtk_widget_realize(window);
	bg_pixmap = gdk_pixmap_create_from_xpm_d(window->window,NULL,NULL,cube_analyzer_xmms_logo_xpm);
	gdk_window_set_back_pixmap(window->window,bg_pixmap,0);
	gtk_signal_connect(GTK_OBJECT(window),"destroy",GTK_SIGNAL_FUNC(cube_analyzer_destroy_cb),NULL);
	gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &window);
	gtk_widget_set_usize(window, WIDTH, HEIGHT);
	gc = gdk_gc_new(window->window);
	draw_pixmap = gdk_pixmap_new(window->window,WIDTH,HEIGHT,gdk_rgb_get_visual()->depth);
	
	bar = gdk_pixmap_new(window->window,25, HEIGHT, gdk_rgb_get_visual()->depth);
	for(i = 0; i < HEIGHT / 2; i++)
	{
		color.red = 0xFFFF;
		color.green = ((i * 255) / (HEIGHT / 2)) << 8;
		color.blue = 0;
		
		gdk_color_alloc(gdk_colormap_get_system(),&color);
		gdk_gc_set_foreground(gc,&color);
		gdk_draw_line(bar,gc,0,i,24,i);
	}
	for(i = 0; i < HEIGHT / 2; i++)
	{
		color.red = (255 - ((i * 255) / (HEIGHT / 2))) <<8;
		color.green = 0xFFFF;
		color.blue = 0;
		
		gdk_color_alloc(gdk_colormap_get_system(),&color);
		gdk_gc_set_foreground(gc,&color);
		gdk_draw_line(bar,gc,0,i + (HEIGHT / 2),24,i + (HEIGHT / 2));
	}
	yscale = NUM_STEPS / log(256);
	gdk_color_black(gdk_colormap_get_system(),&color);
	gdk_gc_set_foreground(gc,&color);
	
	area = gtk_drawing_area_new();
	gtk_container_add(GTK_CONTAINER(window),area);
	gtk_widget_realize(area);
	gdk_window_set_back_pixmap(area->window,bg_pixmap,0);
	
	gtk_widget_show(area);
	gtk_widget_show(window);
	gdk_window_clear(window->window);
	gdk_window_clear(area->window);
}

static void cube_analyzer_cleanup(void)
{
	if(window)
	{
		gtk_widget_destroy(window);
	}
	if(gc)
	{
		gdk_gc_unref(gc);
		gc = NULL;
	}
	if(bg_pixmap)
	{
		gdk_pixmap_unref(bg_pixmap);
		bg_pixmap = NULL;
	}
	if(draw_pixmap)
	{
		gdk_pixmap_unref(draw_pixmap);
		draw_pixmap = NULL;
	}
	if(bar)
	{
		gdk_pixmap_unref(bar);
		bar = NULL;
	}
}

static gint draw_func(gpointer data)
{
	gint i, bar_drawing_height;
	
	if(!window)
	{
		timeout_tag = 0;
		return FALSE;
	}
	
	GDK_THREADS_ENTER();
	gdk_draw_rectangle(draw_pixmap,gc,TRUE,0,0,WIDTH,HEIGHT);

	for(i = 0; i < NUM_BANDS; i++)
	{
		/*if(bar_heights[i] > 4)
			bar_heights[i] -= 4;
		else
			bar_heights[i] = 0;*/
     	        bar_drawing_height = (gint)(bar_heights[i] * HEIGHT / NUM_STEPS);
	        gdk_draw_pixmap(draw_pixmap,gc,bar, 0,HEIGHT - 1 - bar_drawing_height, 1 + i * (WIDTH / NUM_BANDS), HEIGHT - 1 - bar_drawing_height, (WIDTH / NUM_BANDS) - 1, bar_drawing_height);
		
	}
	gdk_window_clear(area->window);
	GDK_THREADS_LEAVE();

	return TRUE;
}

static gint cube_write_function(gpointer data)
{
	gint i, bar_height, j, k;
	gchar cube_string[126]; // It is filled via strcat, ergo:
	cube_string[0] = '\0';
	gchar permutated_cube_string[126]; // It is filled via direct array access, ergo:
	permutated_cube_string[125] = '\0';
	
	
	if (ttyfile) {
		if ( ++tty_delay_counter > TTY_DELAY_LOOPS) {
			tty_delay_counter = 0;
			
			for(i = 4; i > 0; i--)
				strcpy( cube[i], cube[i-1] );
			// for(i = 0; i < 4; i++)
			//	strcpy( cube[i], cube[i+1] );

			for(i = 0; i < (5 < NUM_BANDS ? 5 : NUM_BANDS); i++) {
				bar_height = bar_heights[i];
				for(j = 0; j < 5; j++) {
					cube[0][i*5+j] = (j < bar_height ? '5' : '0' );
				}
			}

			/* -- some old debug foo -- ;
			 * fprintf(ttyfile, "\n");
			 * for(k = 0; k < 5; k++) {
			 * 	for(i = 0; i < NUM_BANDS; i++) {
			 * 		fprintf(ttyfile, "%d", bar_heights[i]);
			 * 		fprintf(ttyfile, "    ");
			 * 	}
			 * }
			 */

			/* old style without cube_string:
			 * fprintf(ttyfile, "%s%s%s%s%s\n\n", cube[0], cube[1], cube[2], cube[3], cube[4]);
			 * The following is the new method:
			 */
			for(i = 0; i < 5; i++) 
				strcat(cube_string, cube[i]);
			// printf("%s\n", cube_string);
			for(i = 0; i < 125; i++)
				permutated_cube_string[ permutation[i] ] = cube_string[i];
			// printf("%s\n\n", permutated_cube_string);
			fprintf(ttyfile, "%s\n\n", permutated_cube_string);
			fflush(ttyfile);
		}
	}
	return TRUE;
}

static void cube_analyzer_playback_start(void)
{
	if(window)
	{
		gdk_window_set_back_pixmap(area->window,draw_pixmap,0);
		gdk_window_clear(area->window);
	}
	
	/* cube-file open - begin */
	ttyfile = fopen(ttyfile_path, "w");
	// if (!ttyfile)
	// 	return 0;
	/* cube-file open - end */
}


static void cube_analyzer_playback_stop(void)
{
	if(GTK_WIDGET_REALIZED(area))
	{
		gdk_window_set_back_pixmap(area->window,bg_pixmap,0);
		gdk_window_clear(area->window);
	}
	/* cube-file close - begin */
	if (ttyfile)
		fclose(ttyfile);
	ttyfile = NULL;
	/* cube-file close - end */	
}


static void cube_analyzer_render_freq(gint16 data[2][256])
{
	gint i,c;
	gint y;
	
	if(!window)
		return;
	for(i = 0; i < NUM_BANDS; i++)
	{
		for(c = xscale[i], y = 0; c < xscale[i + 1]; c++)
		{
			if(data[0][c] > y)
				y = data[0][c];
		}
		y >>= 7;
		if(y != 0)
		{
   		        y = (gint)(log(y) * yscale);
			if(y > HEIGHT - 1)
				y = HEIGHT - 1;
		}
 
		if(y > bar_heights[i])
			bar_heights[i] = y;
		else if(bar_heights[i] > 4)
			bar_heights[i] -= 4;
		else
			bar_heights[i] = 0;
		// printf( "%d %d, ", y, bar_heights[i]);
	}
	// printf("\n");
	draw_func(NULL);
	cube_write_function(NULL);
	return;			
}

/* configure window - BEGIN */

static void configure_ok_cb(gpointer data)
{
	ConfigFile *cfgfile;

	if (ttyfile_path)
		g_free(ttyfile_path);
	ttyfile_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(path_entry)));

	cfgfile = xmms_cfg_open_default_file();
	if (!cfgfile)
		cfgfile = xmms_cfg_new();

	xmms_cfg_write_string(cfgfile, "cube_analyzer", "ttyfile_path", ttyfile_path);
	xmms_cfg_write_default_file(cfgfile);
	xmms_cfg_free(cfgfile);

	gtk_widget_destroy(configure_win);
}

static void configure_destroy(void)
{
}

static void cube_analyzer_configure(void)
{
	static GtkWidget *path_hbox, *path_label, *configure_vbox;
	static GtkWidget *configure_separator;
	static GtkWidget *configure_bbox, *configure_ok, *configure_cancel;

	if (!configure_win)
	{
		configure_win = gtk_window_new(GTK_WINDOW_DIALOG);

		gtk_signal_connect(GTK_OBJECT(configure_win), "destroy", GTK_SIGNAL_FUNC(configure_destroy), NULL);
		gtk_signal_connect(GTK_OBJECT(configure_win), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &configure_win);
		gtk_window_set_title(GTK_WINDOW(configure_win), _("Cube Analyzer Configuration"));
		gtk_window_set_position(GTK_WINDOW(configure_win), GTK_WIN_POS_MOUSE);

		gtk_container_set_border_width(GTK_CONTAINER(configure_win), 10);

		configure_vbox = gtk_vbox_new(FALSE, 10);
		gtk_container_add(GTK_CONTAINER(configure_win), configure_vbox);

		path_hbox = gtk_hbox_new(FALSE, 5);
		gtk_box_pack_start(GTK_BOX(configure_vbox), path_hbox, FALSE, FALSE, 0);

		path_label = gtk_label_new(_("Device File:"));
		gtk_box_pack_start(GTK_BOX(path_hbox), path_label, FALSE, FALSE, 0);
		gtk_widget_show(path_label);

		path_entry = gtk_entry_new();
		if (ttyfile_path)
			gtk_entry_set_text(GTK_ENTRY(path_entry), ttyfile_path);
		gtk_widget_set_usize(path_entry, 200, -1);
		gtk_box_pack_start(GTK_BOX(path_hbox), path_entry, TRUE, TRUE, 0);
		gtk_widget_show(path_entry);

		gtk_widget_show(path_hbox);

		configure_separator = gtk_hseparator_new();
		gtk_box_pack_start(GTK_BOX(configure_vbox), configure_separator, FALSE, FALSE, 0);
		gtk_widget_show(configure_separator);

		configure_bbox = gtk_hbutton_box_new();
		gtk_button_box_set_layout(GTK_BUTTON_BOX(configure_bbox), GTK_BUTTONBOX_END);
		gtk_button_box_set_spacing(GTK_BUTTON_BOX(configure_bbox), 5);
		gtk_box_pack_start(GTK_BOX(configure_vbox), configure_bbox, FALSE, FALSE, 0);

		configure_ok = gtk_button_new_with_label(_("Ok"));
		gtk_signal_connect(GTK_OBJECT(configure_ok), "clicked", GTK_SIGNAL_FUNC(configure_ok_cb), NULL);
		GTK_WIDGET_SET_FLAGS(configure_ok, GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(configure_bbox), configure_ok, TRUE, TRUE, 0);
		gtk_widget_show(configure_ok);
		gtk_widget_grab_default(configure_ok);

		configure_cancel = gtk_button_new_with_label(_("Cancel"));
		gtk_signal_connect_object(GTK_OBJECT(configure_cancel), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(configure_win));
		GTK_WIDGET_SET_FLAGS(configure_cancel, GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(configure_bbox), configure_cancel, TRUE, TRUE, 0);
		gtk_widget_show(configure_cancel);
		gtk_widget_show(configure_bbox);
		gtk_widget_show(configure_vbox);
		gtk_widget_show(configure_win);
	}
}
/* configure window - END */

/* about window - BEGIN */
void cube_analyzer_about(void)
{
	static GtkWidget *dialog;
	
	if (dialog != NULL)
		return;
	
	dialog = xmms_show_message(
		_("About Cube Analyzer"),
		_("Cube Analyzer\n\n"
		  "This is a little plug-in that displays a spectrum analyzer on the LED\n"
		  "Cube that can be connected to the serial port of your computer.\n"
		  "\n"
		  "Have a look at http://www.entropia.de/wiki/LED-Cube\n"
		  "\n"
		  /*
		  "It is based on the code of the sanalyzer visualization plug-in\n"
		  "that comes with the XMMS source code package; some code from\n"
                  "other plug-ins of XMMS may shamelessly be taken as well.\n"
		  */
		  "It is based on the sanalyzer visualization plug-in that comes\n"
		  "with the XMMS source code package; some code from\n"
                  "other plug-ins of XMMS may shamelessly be taken as well.\n"
		  "\n"
		  "Hacked by Max-Gerd Retzlaff <m.retzlaff@gmx.net>.\n"
		  "Have a look at my homepage: http://bl0rg.net/~mgr/\n"
			), _("Ok"), FALSE, NULL, NULL);
	gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
			   GTK_SIGNAL_FUNC(gtk_widget_destroyed),
			   &dialog);
}
/* about window - END */

